// Copyright (c) 2019, the EOMarch project authors. Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'package:flutter/material.dart';

import '../typeahead_base.dart';

/// Simple [Typeahead] based on [TextField]
class Typeahead<E> extends StatefulWidget {
  /// Creates a simple [Typeahead].
  const Typeahead({
    Key key,
    @required this.itemBuilder,
    @required this.suggestsBuilder,
    this.onSuggestTap,
    this.initialData,
    this.validator,
    this.itemToQueryConverter,
    this.delay,
    this.backgroundColor,
    this.autofocus,
    this.enabled,
    this.focusNode,
    this.decoration = const InputDecoration(),
  }) : super(key: key);

  /// Whether this text field should focus itself if nothing else is already
  /// focused.
  ///
  /// If true, the keyboard will open as soon as this text field obtains focus.
  /// Otherwise, the keyboard is only shown after the user taps the text field.
  ///
  /// Defaults to false. Cannot be null.
  final bool autofocus;

  /// If false the text field is "disabled": it ignores taps and its
  /// [decoration] is rendered in grey.
  ///
  /// If non-null this property overrides the [decoration]'s
  /// [Decoration.enabled] property.
  final bool enabled;

  /// Defines the keyboard focus for this widget.
  ///
  /// The [focusNode] is a long-lived object that's typically managed by a
  /// [StatefulWidget] parent. See [FocusNode] for more information.
  ///
  /// To give the keyboard focus to this widget, provide a [focusNode] and then
  /// use the current [FocusScope] to request the focus:
  ///
  /// ```dart
  /// FocusScope.of(context).requestFocus(myFocusNode);
  /// ```
  ///
  /// This happens automatically when the widget is tapped.
  ///
  /// To be notified when the widget gains or loses the focus, add a listener
  /// to the [focusNode]:
  ///
  /// ```dart
  /// focusNode.addListener(() { print(myFocusNode.hasFocus); });
  /// ```
  ///
  /// If null, this widget will create its own [FocusNode].
  ///
  /// ## Keyboard
  ///
  /// Requesting the focus will typically cause the keyboard to be shown
  /// if it's not showing already.
  ///
  /// On Android, the user can hide the keyboard - without changing the focus -
  /// with the system back button. They can restore the keyboard's visibility
  /// by tapping on a text field.  The user might hide the keyboard and
  /// switch to a physical keyboard, or they might just need to get it
  /// out of the way for a moment, to expose something it's
  /// obscuring. In this case requesting the focus again will not
  /// cause the focus to change, and will not make the keyboard visible.
  ///
  /// This widget builds an [EditableText] and will ensure that the keyboard is
  /// showing when it is tapped by calling [EditableTextState.requestKeyboard()].
  final FocusNode focusNode;

  /// The decoration to show around the text field.
  ///
  /// By default, draws a horizontal line under the text field but can be
  /// configured to show an icon, label, hint text, and error text.
  ///
  /// Specify null to remove the decoration entirely (including the
  /// extra padding introduced by the decoration to save space for the labels).
  final InputDecoration decoration;

  /// Defines backroung color of overlay
  ///
  /// If value is ommited, it defaults to [Colors.white]
  final Color backgroundColor;

  /// Initial value of input
  ///
  /// If specified, it will be added to an input via
  /// [itemToQueryConverter] conversion method
  final E initialData;

  /// Defines the amount of time that must ellapse
  /// before [suggestsBuilder] will be called
  ///
  /// If ommited, it defaults to `300 milliseconds`
  final Duration delay;

  /// TODO
  final Validator validator;

  /// TODO
  final ItemBuilder<E> itemBuilder;

  /// TODO
  final SuggestTap<E> onSuggestTap;

  /// TODO
  final SuggestsBuilder<E> suggestsBuilder;

  /// TODO
  final ItemConvert<E> itemToQueryConverter;

  @override
  _TypeaheadState<E> createState() => _TypeaheadState<E>();
}

class _TypeaheadState<E> extends State<Typeahead<E>> {
  @override
  Widget build(BuildContext context) {
    return TypeaheadBase(
      inputBuilder: (context, controller) => TextField(
        controller: controller,
        autofocus: widget.autofocus,
        enabled: widget.enabled,
        focusNode: widget.focusNode,
        decoration: widget.decoration,
      ),
      initialData: widget.initialData,
      itemBuilder: widget.itemBuilder,
      onSuggestTap: widget.onSuggestTap,
      suggestsBuilder: widget.suggestsBuilder,
      itemToQueryConverter: widget.itemToQueryConverter,
      validator: widget.validator,
      delay: widget.delay,
      backgroundColor: widget.backgroundColor,
    );
  }
}
