// Copyright (c) 2019, the EOMarch project authors. Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'dart:async' show FutureOr, StreamController;
import 'dart:math' as math show min;

import 'package:flutter/material.dart';

import 'package:flinq/flinq.dart';
import 'package:rxdart/rxdart.dart';

import 'utils//validators.dart';
import 'utils/keyboard_helper/keyboard_helper_stub.dart'
    if (dart.library.html) 'utils/keyboard_helper/keyboard_helper_web.dart'
    if (dart.library.io) 'utils/keyboard_helper/keyboard_helper_mobile.dart';

const _kProgressIndicatorHeight = 64.0;

/// The signature of validation function from [Validators]
typedef bool Validator(String item);

/// The signature of conversion from `selected` item to a string that represents `search query`
typedef String ItemConvert<E>(E item);

/// The signature of `query results` builder
typedef FutureOr<List<E>> SuggestsBuilder<E>(String query);

/// The signature of `result item` [Widget] builder
typedef Widget ItemBuilder<E>(BuildContext context, E item);

/// The signature of `result item` tap
typedef void SuggestTap<E>(E item, {TextEditingController controller});

/// The signature of input [Widget] builder
typedef Widget InputBuilder(
    BuildContext context, TextEditingController controller);

/// Base typeahead widget
class TypeaheadBase<E> extends StatefulWidget {
  /// Creates a base typeahead.
  const TypeaheadBase({
    Key key,
    @required this.itemBuilder,
    @required this.inputBuilder,
    @required this.suggestsBuilder,
    this.onSuggestTap,
    this.validator,
    this.initialData,
    this.itemToQueryConverter,
    this.delay = const Duration(milliseconds: 300),
    this.backgroundColor = Colors.white,
  }) : super(key: key);

  /// Defines backroung color of overlay
  ///
  /// If value is ommited, it defaults to [Colors.white]
  final Color backgroundColor;

  /// Initial value of input
  ///
  /// If specified, it will be added to an input via
  /// [itemToQueryConverter] conversion method
  final E initialData;

  /// Defines the amount of time that must ellapse
  /// before [suggestsBuilder] will be called
  ///
  /// If ommited, it defaults to `300 milliseconds`
  final Duration delay;

  /// TODO
  final Validator validator;

  /// TODO
  final ItemBuilder<E> itemBuilder;

  /// TODO
  final SuggestTap<E> onSuggestTap;

  /// TODO
  final SuggestsBuilder<E> suggestsBuilder;

  /// TODO
  final ItemConvert<E> itemToQueryConverter;

  /// TODO
  final InputBuilder inputBuilder;

  @override
  TypeaheadBaseState<E> createState() => TypeaheadBaseState<E>();
}

/// State of base typeahead
class TypeaheadBaseState<E> extends State<TypeaheadBase<E>> {
  final _debounceController = StreamController<String>.broadcast();
  final _suggestController = StreamController<List<E>>.broadcast();
  final _layerLink = LayerLink();

  TextEditingController _controller;
  OverlayEntry _overlayEntry;
  String _prevQuery = '';

  /// This method is used to remove overlay from screen
  @protected
  void removeOverlay() {
    _overlayEntry?.remove();
    _overlayEntry = null;
  }

  bool _validator(String text) =>
      widget.validator != null ? widget.validator(text) : Validators().notEmpty;

  Widget _itemBuilder(E elem) => ListTile(
        title: widget.itemBuilder(context, elem),
        onTap: () {
          if (widget.itemToQueryConverter != null) {
            _controller.text = widget.itemToQueryConverter(elem);
            _controller.selection =
                TextSelection.collapsed(offset: _controller.text.length);
          }
          _prevQuery = _controller.text;
          widget.onSuggestTap(elem, controller: _controller);
          removeOverlay();
        },
      );

  /// This method is used to specify how should
  /// overlay appear
  @protected
  OverlayEntry createOverlay() => OverlayEntry(
        builder: (_) {
          final RenderBox renderBox = context.findRenderObject();
          final size = renderBox.size;

          final halfScreenSize = MediaQuery.of(context).size.height / 2;
          final availableSpace = MediaQuery.of(context).size.height -
              (renderBox.localToGlobal(Offset.zero)?.dy ?? .0) +
              (size?.height ?? .0);

          final maxSize = math.min(halfScreenSize, availableSpace);
          final offset = Offset(0.0, size.height + 5.0);

          return Positioned(
            width: size.width,
            child: Transform.translate(
              offset: offset,
              child: CompositedTransformFollower(
                link: _layerLink,
                showWhenUnlinked: false,
                offset: offset,
                child: Material(
                  elevation: 4.0,
                  color: widget.backgroundColor ?? Colors.transparent,
                  child: ConstrainedBox(
                    constraints: BoxConstraints(
                      maxHeight: maxSize,
                      minHeight: _kProgressIndicatorHeight,
                      maxWidth: size.width,
                      minWidth: size.width,
                    ),
                    child: StreamBuilder<List<E>>(
                      stream: _suggestController.stream,
                      builder: (context, snapshot) {
                        // TODO: add check for null date (should consider it as error)

                        final suggests = snapshot.data;

                        return AnimatedCrossFade(
                          duration: kThemeAnimationDuration,
                          crossFadeState: suggests != null
                              ? CrossFadeState.showSecond
                              : CrossFadeState.showFirst,
                          firstChild: Container(
                            width: double.infinity,
                            height: _kProgressIndicatorHeight,
                            child: Center(child: CircularProgressIndicator()),
                          ),
                          secondChild: Scrollbar(
                            child: ListView(
                              padding: EdgeInsets.zero,
                              shrinkWrap: true,
                              children:
                                  suggests?.mapList((_) => _itemBuilder(_)) ??
                                      <Widget>[],
                            ),
                          ),
                        );
                      },
                    ),
                  ),
                ),
              ),
            ),
          );
        },
      );

  @override
  void initState() {
    super.initState();
    _controller = TextEditingController(
        text: widget.itemToQueryConverter(widget.initialData));

    _controller.addListener(() => _debounceController.add(_controller.text));

    _debounceController.stream
        .transform(
            DebounceStreamTransformer((_) => TimerStream(_, widget.delay)))
        .listen((_) async {
      if (_prevQuery != _controller.text &&
          _validator(_controller.text) == true) {
        _prevQuery = _controller.text;

        if (_overlayEntry == null) {
          _overlayEntry = createOverlay();
          Overlay.of(context).insert(_overlayEntry);
        }

        final suggests = await widget.suggestsBuilder(_controller.text);
        _suggestController.add(suggests);
      }
    });

    KeyboardHelper.keys.listen((_) {
      if (_ == 'Escape') {
        removeOverlay();
      }
    });
  }

  @override
  Widget build(BuildContext context) {
    // TODO: add WillPopScope for intercepting mobile back, and close overlay
    return CompositedTransformTarget(
      link: _layerLink,
      child: Material(
        color: Colors.transparent,
        child: widget.inputBuilder(context, _controller),
      ),
    );
  }

  @override
  void dispose() {
    super.dispose();
    _debounceController.close();
    _suggestController.close();
    _controller.dispose();
    removeOverlay();
  }
}
